# MODULES
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, Union

# CONNTEXTLIB
from contextlib import AbstractContextManager

# SQLALCHEMY
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session, InstrumentedAttribute, Query

# UTILS
from session_repository.utils import (
    _FilterType,
    apply_no_load,
    apply_filters,
    apply_order_by,
    apply_limit,
    apply_pagination,
)


def with_session(func):
    def wrapper(self, *args, **kwargs):
        if kwargs.get("current_session") is not None:
            return func(self, *args, **kwargs)

        with self.session_manager() as session:
            kwargs["current_session"] = session

            return func(self, *args, **kwargs)

    return wrapper


T = TypeVar("T", bound=declarative_base())


class SessionRepository:
    def __init__(
        self,
        session_factory: Callable[..., AbstractContextManager[Session]],
    ) -> None:
        self._session_factory = session_factory

    def session_manager(self):
        return self._session_factory()

    @with_session
    def _select(
        self,
        model: Type[T],
        filters: Optional[_FilterType] = None,
        optional_filters: Optional[_FilterType] = None,
        disabled_relationships: Optional[Dict[InstrumentedAttribute, Any]] = None,
        current_session: Optional[Session] = None,
    ) -> Optional[T]:
        query = current_session.query(model)

        return self._select_query(
            query=query,
            filters=filters,
            optional_filters=optional_filters,
            disabled_relationships=disabled_relationships,
        )

    def _select_query(
        self,
        query: Query,
        filters: Optional[_FilterType] = None,
        optional_filters: Optional[_FilterType] = None,
        disabled_relationships: Optional[Dict[InstrumentedAttribute, Any]] = None,
    ) -> Optional[Any]:
        query = apply_no_load(
            query=query,
            relationship_dict=disabled_relationships,
        )
        query = apply_filters(
            query=query,
            filter_dict=filters,
        )
        query = apply_filters(
            query=query,
            filter_dict=optional_filters,
            with_optional=True,
        )
        result = query.first()

        return result

    @with_session
    def _select_all(
        self,
        model: Type[T],
        filters: Optional[_FilterType] = None,
        optional_filters: Optional[_FilterType] = None,
        disabled_relationships: Optional[Dict[InstrumentedAttribute, Any]] = None,
        order_by: Optional[Union[List[str], str]] = None,
        direction: Optional[str] = None,
        limit: int = None,
        current_session: Optional[Session] = None,
    ) -> List[T]:
        query = current_session.query(model)

        return self._select_all_query(
            query=query,
            model=model,
            filters=filters,
            optional_filters=optional_filters,
            disabled_relationships=disabled_relationships,
            order_by=order_by,
            direction=direction,
            limit=limit,
        )

    def _select_all_query(
        self,
        query: Query,
        model: Type[T],
        filters: Optional[_FilterType] = None,
        optional_filters: Optional[_FilterType] = None,
        disabled_relationships: Optional[Dict[InstrumentedAttribute, Any]] = None,
        order_by: Optional[Union[List[str], str]] = None,
        direction: Optional[str] = None,
        limit: int = None,
    ) -> List[T]:
        query = apply_no_load(
            query=query,
            relationship_dict=disabled_relationships,
        )
        query = apply_filters(
            query=query,
            filter_dict=filters,
        )
        query = apply_filters(
            query=query,
            filter_dict=optional_filters,
            with_optional=True,
        )
        query = apply_order_by(
            query=query,
            model=model,
            order_by=order_by,
            direction=direction,
        )
        query = apply_limit(
            query=query,
            limit=limit,
        )

        results = query.all()

        return results

    @with_session
    def _select_paginate(
        self,
        model: Type[T],
        page: int,
        per_page: int,
        filters: Optional[_FilterType] = None,
        optional_filters: Optional[_FilterType] = None,
        disabled_relationships: Optional[Dict[InstrumentedAttribute, Any]] = None,
        order_by: Optional[Union[List[str], str]] = None,
        direction: Optional[str] = None,
        limit: int = None,
        current_session: Optional[Session] = None,
    ) -> Tuple[List[T], str]:
        query = current_session.query(model)

        return self._select_paginate_query(
            query=query,
            model=model,
            page=page,
            per_page=per_page,
            filters=filters,
            optional_filters=optional_filters,
            disabled_relationships=disabled_relationships,
            order_by=order_by,
            direction=direction,
            limit=limit,
        )

    def _select_paginate_query(
        self,
        query: Query,
        model: Type[T],
        page: int,
        per_page: int,
        filters: Optional[_FilterType] = None,
        optional_filters: Optional[_FilterType] = None,
        disabled_relationships: Optional[Dict[InstrumentedAttribute, Any]] = None,
        order_by: Optional[Union[List[str], str]] = None,
        direction: Optional[str] = None,
        limit: int = None,
    ) -> Tuple[List[T], str]:
        query = apply_no_load(
            query=query,
            relationship_dict=disabled_relationships,
        )
        query = apply_filters(
            query=query,
            filter_dict=filters,
        )
        query = apply_filters(
            query=query,
            filter_dict=optional_filters,
            with_optional=True,
        )
        query = apply_order_by(
            query=query,
            model=model,
            order_by=order_by,
            direction=direction,
        )
        query = apply_limit(
            query=query,
            limit=limit,
        )
        query, pagination = apply_pagination(
            query=query,
            page=page,
            per_page=per_page,
        )

        results = query.all()

        return results, pagination

    @with_session
    def _update_all(
        self,
        model: Type[T],
        values: Dict,
        filters: Optional[_FilterType] = None,
        flush: bool = False,
        commit: bool = False,
        current_session: Optional[Session] = None,
    ) -> List[T]:
        rows = self._select_all(
            model=model,
            filters=filters,
            current_session=current_session,
        )

        if len(rows) == 0:
            return rows

        for row in rows:
            for key, value in values.items():
                setattr(row, key, value)

        if flush:
            current_session.flush()
        if commit:
            current_session.commit()

        [current_session.refresh(row) for row in rows]

        return rows

    @with_session
    def _update(
        self,
        model: Type[T],
        values: Dict,
        filters: Optional[_FilterType] = None,
        flush: bool = False,
        commit: bool = False,
        current_session: Optional[Session] = None,
    ) -> T:
        row = self._select(
            model=model,
            filters=filters,
            current_session=current_session,
        )

        if row is None:
            return

        for key, value in values.items():
            setattr(row, key, value)

        if flush:
            current_session.flush()
        if commit:
            current_session.commit()

        current_session.refresh(row)

        return row

    @with_session
    def _add_all(
        self,
        data: List[T],
        flush: bool = False,
        commit: bool = False,
        current_session: Optional[Session] = None,
    ) -> List[T]:
        current_session.add_all(data)
        if flush:
            current_session.flush()
        if commit:
            current_session.commit()

        if flush or commit:
            [current_session.refresh(item) for item in data]

        return data

    @with_session
    def _add(
        self,
        data: T,
        flush: bool = False,
        commit: bool = False,
        current_session: Optional[Session] = None,
    ) -> T:
        current_session.add(data)
        if flush:
            current_session.flush()
        if commit:
            current_session.commit()

        if flush or commit:
            current_session.refresh(data)

        return data

    @with_session
    def _delete(
        self,
        model: T,
        filters: Optional[_FilterType] = None,
        flush: bool = True,
        commit: bool = False,
        current_session: Optional[Session] = None,
    ) -> bool:
        rows: List = self._select_all(
            model=model,
            filters=filters,
            current_session=current_session,
        )

        if len(rows) == 0:
            return False

        for row in rows:
            current_session.delete(row)

        if flush:
            current_session.flush()
        if commit:
            current_session.commit()

        return True



# MODULES
import enum


class Operators(enum.Enum):
    LIKE = "LIKE"
    NOT_LIKE = "NOT_LIKE"
    ILIKE = "ILIKE"
    NOT_ILIKE = "NOT_ILIKE"
    EQUAL = "EQUAL"
    IEQUAL = "IEQUAL"
    DIFFERENT = "DIFFERENT"
    IDIFFERENT = "IDIFFERENT"
    BETWEEN = "BETWEEN"
    BETWEEN_OR_EQUAL = "BETWEEN_OR_EQUAL"
    SUPERIOR = "SUPERIOR"
    SUPERIOR_OR_EQUAL = "SUPERIOR_OR_EQUAL"
    INFERIOR = "INFERIOR"
    INFERIOR_OR_EQUAL = "INFERIOR_OR_EQUAL"
    IN = "IN"
    IIN = "IIN"
    NOT_IN = "NOT_IN"
    NOT_IIN = "NOT_IIN"
    HAS = "HAS"
    ANY = "ANY"



# MODULES
import json
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Tuple,
    Union,
)

# SQLALCHEMY
from sqlalchemy import and_, asc, desc, tuple_, func, null
from sqlalchemy.orm import (
    Query,
    InstrumentedAttribute,
    noload,
)
from sqlalchemy.sql.elements import Null

# Enum
from session_repository.enum import Operators

_FilterType = Dict[Union[InstrumentedAttribute, Tuple[InstrumentedAttribute]], Any]


def apply_no_load(
    query: Query,
    relationship_dict: Dict[InstrumentedAttribute, Any],
    parents: List[InstrumentedAttribute] = None,
):
    if relationship_dict is None:
        return query

    for relationship, sub_relationships in relationship_dict.items():
        if relationship is None or not isinstance(relationship, InstrumentedAttribute):
            continue

        sub_items = [relationship] if parents is None else [*parents, relationship]
        if sub_relationships is None:
            query = query.options(noload(*sub_items))
        else:
            query = apply_no_load(
                query,
                relationship_dict=sub_relationships,
                parents=sub_items,
            )

    return query


def apply_filters(
    query: Query,
    filter_dict: _FilterType,
    with_optional: bool = False,
):
    filters = get_filters(
        filters=filter_dict,
        with_optional=with_optional,
    )

    return query if len(filters) == 0 else query.filter(and_(*filters))


def apply_order_by(
    query: Query,
    model,
    order_by: Union[List[str], str],
    direction: Union[List[str], str],
):
    if order_by is None or direction is None:
        return query

    if isinstance(order_by, str):
        order_by = [order_by]

    if isinstance(direction, str):
        direction = [direction]

    if len(order_by) != len(direction):
        raise ValueError("order_by length must be equals to direction length")

    order_by_list = []
    for column, dir in zip(order_by, direction):
        if dir == "desc":
            order_by_list.append(desc(getattr(model, column)))
        elif dir == "asc":
            order_by_list.append(asc(getattr(model, column)))

    return query.order_by(*order_by_list)


def build_order_by(
    model,
    order_by: dict,
):
    if isinstance(order_by, dict):
        order_by_list = []
        for key, value in order_by.items():
            if isinstance(value, dict):
                relationship = getattr(model, key)
                order_by_relationship = build_order_by(value, relationship)
                order_by_list.extend(order_by_relationship)
            else:
                column = getattr(model, key)
                if value == "ASC":
                    order_by_list.append(asc(column))
                elif value == "DESC":
                    order_by_list.append(desc(column))
        return order_by_list
    else:
        raise ValueError("Invalid nomenclature format.")


def apply_pagination(
    query: Query,
    page: int,
    per_page: int,
):
    pagination = None
    if page is not None and per_page is not None:
        total_results = query.count()
        total_pages = (total_results + per_page - 1) // per_page

        pagination = {
            "total": total_results,
            "page": page,
            "per_page": per_page,
            "total_pages": total_pages,
        }

        pagination = json.dumps(pagination)

        query = query.offset((page - 1) * per_page).limit(per_page)

    return query, pagination


def apply_limit(
    query: Query,
    limit: int,
):
    return query.limit(limit) if limit is not None else query


def get_conditions_from_dict(
    values: _FilterType,
    with_optional: bool = False,
):
    conditions = []
    for key, value in values.items():
        if type(value) == set:
            value = list(value)
        elif type(value) == dict:
            for k, v in value.items():
                if with_optional and v is None:
                    continue

                match k:
                    case Operators.EQUAL:
                        conditions.append(key == v)
                    case Operators.IEQUAL:
                        if not isinstance(v, Null):
                            conditions.append(func.lower(key) == func.lower(v))
                        else:
                            conditions.append(key == v)
                    case Operators.DIFFERENT:
                        conditions.append(key != v)
                    case Operators.IDIFFERENT:
                        if not isinstance(v, Null):
                            conditions.append(func.lower(key) != func.lower(v))
                        else:
                            conditions.append(key != v)
                    case Operators.LIKE:
                        if not isinstance(v, Null):
                            conditions.append(key.like(v))
                        else:
                            conditions.append(key == v)
                    case Operators.NOT_LIKE:
                        if not isinstance(v, Null):
                            conditions.append(~key.like(v))
                        else:
                            conditions.append(key != v)
                    case Operators.ILIKE:
                        if not isinstance(v, Null):
                            conditions.append(key.ilike(v))
                        else:
                            conditions.append(key == v)
                    case Operators.NOT_ILIKE:
                        if not isinstance(v, Null):
                            conditions.append(~key.ilike(v))
                        else:
                            conditions.append(key != v)
                    case Operators.BETWEEN:
                        if len(v) != 2:
                            continue
                        if v[0] is not None:
                            conditions.append(key > v[0])
                        if v[1] is not None:
                            conditions.append(key < v[1])
                    case Operators.BETWEEN_OR_EQUAL:
                        if len(v) != 2:
                            continue
                        if v[0] is not None:
                            conditions.append(key >= v[0])
                        if v[1] is not None:
                            conditions.append(key <= v[1])
                    case Operators.SUPERIOR:
                        conditions.append(key > v)
                    case Operators.INFERIOR:
                        conditions.append(key < v)
                    case Operators.SUPERIOR_OR_EQUAL:
                        conditions.append(key >= v)
                    case Operators.INFERIOR_OR_EQUAL:
                        conditions.append(key <= v)
                    case Operators.IN:
                        v = v if isinstance(v, Iterable) else [v]
                        if isinstance(key, tuple):
                            conditions.append(tuple_(*key).in_(v))
                        else:
                            conditions.append(key.in_(v))
                    case Operators.IIN:
                        v = v if isinstance(v, Iterable) else [v]
                        if isinstance(key, tuple):
                            conditions.append(
                                tuple_([func.lower(key_) for key_ in key]).in_(
                                    [
                                        func.lower(v_)
                                        if not isinstance(v_, Null)
                                        else v_
                                        for v_ in v
                                    ]
                                )
                            )
                        else:
                            conditions.append(
                                func.lower(key).in_(
                                    [
                                        func.lower(v_)
                                        if not isinstance(v_, Null)
                                        else v_
                                        for v_ in v
                                    ]
                                )
                            )
                    case Operators.NOT_IN:
                        v = v if isinstance(v, Iterable) else [v]
                        if isinstance(key, tuple):
                            conditions.append(tuple_(*key).notin_(v))
                        else:
                            conditions.append(key.notin_(v))

                    case Operators.NOT_IIN:
                        v = v if isinstance(v, Iterable) else [v]
                        if isinstance(key, tuple):
                            conditions.append(
                                tuple_([func.lower(key_) for key_ in key]).notin_(
                                    [
                                        func.lower(v_)
                                        if not isinstance(v_, Null)
                                        else v_
                                        for v_ in v
                                    ]
                                )
                            )
                        else:
                            conditions.append(
                                func.lower(key).notin_(
                                    [
                                        func.lower(v_)
                                        if not isinstance(v_, Null)
                                        else v_
                                        for v_ in v
                                    ]
                                )
                            )
                    case Operators.HAS:
                        v = get_filters(
                            v,
                            with_optional=with_optional,
                        )
                        for condition in v:
                            conditions.append(key.has(condition))
                    case Operators.ANY:
                        v = get_filters(
                            v,
                            with_optional=with_optional,
                        )

                        if len(v) == 0:
                            continue

                        conditions.append(key.any(and_(*v)))

    return conditions


def get_filters(
    filters: _FilterType,
    with_optional: bool = False,
):
    if filters is None:
        return []
    if not isinstance(filters, dict):
        raise TypeError("<filters> must be type of <dict>")

    filters = [{x: y} for x, y in filters.items()]

    conditions = []
    for filter_c in filters:
        if not type(filter_c) == dict:
            continue

        conditions_from_dict = get_conditions_from_dict(
            filter_c,
            with_optional=with_optional,
        )
        conditions.extend(conditions_from_dict)

    return conditions
